/*
 * gss-server.c --- originally taken from the Krb5 GSS Sample
 * application.
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <ctype.h>

#include <gssapi/gssapi_generic.h>
#include "gss-misc.h"
#include "randpass.h"
#include "kludge.h"

#define LOG_FILE "/usr/tmp/kludge.log"
#define SECRET_FILE "/tmp/secret"

void usage()
{
     fprintf(stderr, "Usage: gss-server [-port port] \n");
     fprintf(stderr, "       [-inetd] [-logfile file] [service_name]\n");
     exit(1);
}

FILE *log;

/*
 * Function: server_acquire_creds
 *
 * Purpose: imports a service name and acquires credentials for it
 *
 * Arguments:
 *
 * 	service_name	(r) the ASCII service name
 * 	server_creds	(w) the GSS-API service credentials
 *
 * Returns: 0 on success, -1 on failure
 *
 * Effects:
 *
 * The service name is imported with gss_import_name, and service
 * credentials are acquired with gss_acquire_cred.  If either opertion
 * fails, an error message is displayed and -1 is returned; otherwise,
 * 0 is returned.
 */
int server_acquire_creds(service_name, server_creds)
     char *service_name;
     gss_cred_id_t *server_creds;
{
     gss_buffer_desc name_buf;
     gss_name_t server_name;
     OM_uint32 maj_stat, min_stat;

     name_buf.value = service_name;
     name_buf.length = strlen(name_buf.value) + 1;
     maj_stat = gss_import_name(&min_stat, &name_buf, 
				(gss_OID) gss_nt_service_name, &server_name);
     if (maj_stat != GSS_S_COMPLETE) {
	  display_status("importing name", maj_stat, min_stat);
	  return -1;
     }

     maj_stat = gss_acquire_cred(&min_stat, server_name, 0,
				 GSS_C_NULL_OID_SET, GSS_C_ACCEPT,
				 server_creds, NULL, NULL);
     if (maj_stat != GSS_S_COMPLETE) {
	  display_status("acquiring credentials", maj_stat, min_stat);
	  return -1;
     }

     (void) gss_release_name(&min_stat, &server_name);

     return 0;
}

/*
 * Function: server_establish_context
 *
 * Purpose: establishses a GSS-API context as a specified service with
 * an incoming client, and returns the context handle and associated
 * client name
 *
 * Arguments:
 *
 * 	s		(r) an established TCP connection to the client
 * 	service_creds	(r) server credentials, from gss_acquire_cred
 * 	context		(w) the established GSS-API context
 * 	client_name	(w) the client's ASCII name
 *
 * Returns: 0 on success, -1 on failure
 *
 * Effects:
 *
 * Any valid client request is accepted.  If a context is established,
 * its handle is returned in context and the client name is returned
 * in client_name and 0 is returned.  If unsuccessful, an error
 * message is displayed and -1 is returned.
 */
int server_establish_context(s, server_creds, context, client_name, ret_flags)
     int s;
     gss_cred_id_t server_creds;
     gss_ctx_id_t *context;
     gss_buffer_t client_name;
     OM_uint32 *ret_flags;
{
     gss_buffer_desc send_tok, recv_tok;
     gss_name_t client;
     gss_OID doid;
     OM_uint32 maj_stat, min_stat;

     *context = GSS_C_NO_CONTEXT;
     
     do {
	  if (recv_token(s, &recv_tok) < 0)
	       return -1;

	  maj_stat =
	       gss_accept_sec_context(&min_stat,
				      context,
				      server_creds,
				      &recv_tok,
				      GSS_C_NO_CHANNEL_BINDINGS,
				      &client,
				      &doid,
				      &send_tok,
				      ret_flags,
				      NULL, 	/* ignore time_rec */
				      NULL); 	/* ignore del_cred_handle */

	  if (maj_stat!=GSS_S_COMPLETE && maj_stat!=GSS_S_CONTINUE_NEEDED) {
	       display_status("accepting context", maj_stat, min_stat);
	       (void) gss_release_buffer(&min_stat, &recv_tok);
	       return -1;
	  }

	  (void) gss_release_buffer(&min_stat, &recv_tok);

	  if (send_tok.length != 0) {
	       if (send_token(s, &send_tok) < 0) {
		    fprintf(log, "failure sending token\n");
		    return -1;
	       }

	       (void) gss_release_buffer(&min_stat, &send_tok);
	  }
     } while (maj_stat == GSS_S_CONTINUE_NEEDED);

     maj_stat = gss_display_name(&min_stat, client, client_name, &doid);
     if (maj_stat != GSS_S_COMPLETE) {
	  display_status("displaying name", maj_stat, min_stat);
	  return -1;
     }
     maj_stat = gss_release_name(&min_stat, &client);
     if (maj_stat != GSS_S_COMPLETE) {
	  display_status("releasing name", maj_stat, min_stat);
	  return -1;
     }
     return 0;
}

/*
 * Function: create_socket
 *
 * Purpose: Opens a listening TCP socket.
 *
 * Arguments:
 *
 * 	port		(r) the port number on which to listen
 *
 * Returns: the listening socket file descriptor, or -1 on failure
 *
 * Effects:
 *
 * A listening socket on the specified port and created and returned.
 * On error, an error message is displayed and -1 is returned.
 */
int create_socket(port)
     u_short port;
{
     struct sockaddr_in saddr;
     int s;
     int on = 1;
     
     saddr.sin_family = AF_INET;
     saddr.sin_port = htons(port);
     saddr.sin_addr.s_addr = INADDR_ANY;

     if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  perror("creating socket");
	  return -1;
     }
     /* Let the socket be reused right away */
     (void) setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
     if (bind(s, (struct sockaddr *) &saddr, sizeof(saddr)) < 0) {
	  perror("binding socket");
	  (void) close(s);
	  return -1;
     }
     if (listen(s, 5) < 0) {
	  perror("listening on socket");
	  (void) close(s);
	  return -1;
     }
     return s;
}

/*
 * Initialize the kludge server
 */
int init_kludge_server(app_ctx, service_name)
	char		*service_name;
	kludge_t	app_ctx;
{
	app_ctx->listen_socket = 0;
	app_ctx->socket = 0;

	app_ctx->log = fopen(LOG_FILE, "a");
	if (!app_ctx->log) {
		perror(LOG_FILE);
		return -1;
	}

	randpass_init(SECRET_FILE);

	if (server_acquire_creds(service_name, &app_ctx->server_creds) < 0)
		return -1;

	return 0;
}

/*
 * shutdown_kludge server
 */
void shutdown_kludge_server(app_ctx)
	kludge_t	app_ctx;
{
	OM_uint32 min_stat;
	
	(void) gss_release_cred(&min_stat, &app_ctx->server_creds);
}

/*
 * Function: kludge_server
 *
 * Purpose: Performs the "kludge" service.
 *
 * Arguments:
 *
 * 	s		(r) a TCP socket on which a connection has been
 *			accept()ed
 * 	service_name	(r) the ASCII name of the GSS-API service to
 * 			establish a context as
 * 
 * Returns: -1 on error
 *
 * Effects:
 *
 * sign_server establishes a context, and performs a single sign request.
 *
 * A sign request is a single GSS-API sealed token.  The token is
 * unsealed and a signature block, produced with gss_sign, is returned
 * to the sender.  The context is the destroyed and the connection
 * closed.
 *
 * If any error occurs, -1 is returned.
 */
int kludge_server(app_ctx)
	kludge_t	app_ctx;
{
	gss_buffer_desc client_name, msg_buf, xmit_buf;
	gss_ctx_id_t context;
	OM_uint32 maj_stat, min_stat;
	int conf_state, ret_flags, retval;
	char	password[9], *username, *cp;
     
	/* Establish a context with the client */
	if (server_establish_context(app_ctx->socket,
				     app_ctx->server_creds, &context,
				     &client_name, &ret_flags) < 0)
		return(-1);
	username = malloc(client_name.length+1);
	if (!username) {
		printf("No memory to allocate client name\n");
		return -1;
	}
	strncpy(username, client_name.value, client_name.length);
	username[client_name.length] = '\0';
	cp = strchr(username, '@');
	if (cp)
		*cp = 0;

	(void) gss_release_buffer(&min_stat, &client_name);

	retval = recv_sealed_token(app_ctx->socket, context, &msg_buf,
				   &conf_state, &maj_stat, &min_stat);
	if (retval) {
		printf("Couldn't get token: %s\n", strerror(retval));
		return(-1);
	}
	if (maj_stat != GSS_S_COMPLETE) {
		display_status("unsealing initial message", maj_stat,
			       min_stat);
		return(-1);
	}
	if (!conf_state) {
		fprintf(stderr, "Warning!  Message not encrypted.\n");
	}

	/*
	 * XXX check version code
	 */

	randpass_gen(password, 8);
	printf("Set password: '%s' for user '%s'\n", password, username);

	retval = create_kludge_token(username, password, &xmit_buf);
	if (retval) {
		printf("Error creating kludge token: %d\n", retval);
		return(-1);
	}
		
	conf_state = 1;
	retval = send_sealed_token(app_ctx->socket, context,
				   &xmit_buf, &conf_state,
				   &maj_stat, &min_stat);
	if (retval) {
		printf("Couldn't send token: %s\n", strerror(retval));
		return(-1);
	}
	if (maj_stat != GSS_S_COMPLETE) {
		display_status("sealing initial message", maj_stat,
			       min_stat);
		return(-1);
	}
	if (!conf_state) {
		fprintf(stderr, "Warning!  Message not sent encrypted.\n");
	}
	
	/* Delete context */
	maj_stat = gss_delete_sec_context(&min_stat, &context, NULL);
	if (maj_stat != GSS_S_COMPLETE) {
		display_status("deleting context", maj_stat, min_stat);
		return(-1);
	}

	fflush(log);
	free(username);
	free(xmit_buf.value);

	return(0);
}

int
main(argc, argv)
	int argc;
	char **argv;
{
	char *service_name;
	u_short port = 17007;
	int s = 0;
	int once = 0;
	int do_inetd = 0;
	struct kludge appl_ctx;
	
	log = stdout;
	display_file = stdout;
	argc--; argv++;
	while (argc) {
		if (strcmp(*argv, "-port") == 0) {
			argc--; argv++;
			if (!argc) usage();
			port = atoi(*argv);
		} else if (strcmp(*argv, "-once") == 0) {
			once = 1;
		} else if (strcmp(*argv, "-inetd") == 0) {
			do_inetd = 1;
		} else if (strcmp(*argv, "-logfile") == 0) {
			argc--; argv++;
			if (!argc) usage();
			log = fopen(*argv, "a");
			display_file = log;
			if (!log) {
				perror(*argv);
				exit(1);
			}
		} else
			break;
		argc--; argv++;
	}
	if (argc != 1)
		usage();

	if ((*argv)[0] == '-')
		usage();
	
	service_name = *argv;

	if (init_kludge_server(&appl_ctx, service_name)  < 0)
		exit(1);

	if (do_inetd) {
		close(1);
		close(2);

		kludge_server(&appl_ctx);
		close(0);
	} else {
		int stmp;

		if ((stmp = create_socket(port))) {
			do {
				/* Accept a TCP connection */
				s = accept(stmp, NULL, 0);
				if (s < 0) {
					perror("accepting connection");
					continue;
				}
				appl_ctx.socket = s;
				/*
				 * this return value is not
				 * checked, because there's
				 * not really anything to do
				 * if it fails
				 */
				kludge_server(&appl_ctx);
			} while (!once);
		}
		close(stmp);
	}
	
	shutdown_kludge_server(&appl_ctx);

	/*NOTREACHED*/
	if (s)
		(void) close(s);
	return 0;
}
